package photo

import (
	"net/http"
	"strconv"

	"code.nonshy.com/nonshy/website/pkg/log"
	"code.nonshy.com/nonshy/website/pkg/models"
	"code.nonshy.com/nonshy/website/pkg/session"
	"code.nonshy.com/nonshy/website/pkg/templates"
)

// View photo controller to see the comment thread.
func View() http.HandlerFunc {
	tmpl := templates.Must("photo/permalink.html")
	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		// Required query param: the photo ID.
		var photo *models.Photo
		if idStr := r.FormValue("id"); idStr == "" {
			session.FlashError(w, r, "Missing photo ID parameter.")
			templates.Redirect(w, "/")
			return
		} else {
			if idInt, err := strconv.Atoi(idStr); err != nil {
				session.FlashError(w, r, "Invalid ID parameter.")
				templates.Redirect(w, "/")
				return
			} else {
				if found, err := models.GetPhoto(uint64(idInt)); err != nil {
					templates.NotFoundPage(w, r)
					return
				} else {
					photo = found
				}
			}
		}

		// Find the photo's owner.
		user, err := models.GetUser(photo.UserID)
		if err != nil {
			templates.NotFoundPage(w, r)
			return
		}

		// Load the current user in case they are viewing their own page.
		currentUser, err := session.CurrentUser(r)
		if err != nil {
			session.FlashError(w, r, "Unexpected error: couldn't get CurrentUser")
		}
		var isOwnPhoto = currentUser.ID == user.ID

		// Is either one blocking?
		if models.IsBlocking(currentUser.ID, user.ID) && !currentUser.IsAdmin {
			templates.NotFoundPage(w, r)
			return
		}

		// Is this user private and we're not friends?
		var (
			areFriends = models.AreFriends(user.ID, currentUser.ID)
			isPrivate  = user.Visibility == models.UserVisibilityPrivate && !areFriends
		)
		if isPrivate && !currentUser.IsAdmin && !isOwnPhoto {
			session.FlashError(w, r, "This user's profile page and photo gallery are private.")
			templates.Redirect(w, "/u/"+user.Username)
			return
		}

		// Is this a private photo and are we allowed to see?
		isGranted := models.IsPrivateUnlocked(user.ID, currentUser.ID)
		if photo.Visibility == models.PhotoPrivate && !isGranted && !currentUser.IsAdmin {
			templates.NotFoundPage(w, r)
			return
		}

		// Get Likes information about these photos.
		likeMap := models.MapLikes(currentUser, "photos", []uint64{photo.ID})
		commentMap := models.MapCommentCounts("photos", []uint64{photo.ID})

		// Get all the comments.
		comments, err := models.ListComments("photos", photo.ID)
		if err != nil {
			log.Error("Couldn't list comments for photo %d: %s", photo.ID, err)
		}

		// Get the like map for these comments.
		commentIDs := []uint64{}
		for _, com := range comments {
			commentIDs = append(commentIDs, com.ID)
		}
		commentLikeMap := models.MapLikes(currentUser, "comments", commentIDs)

		// Populate the user relationships in these comments.
		models.SetUserRelationshipsInComments(currentUser, comments)

		// Is the current user subscribed to notifications on this thread?
		_, isSubscribed := models.IsSubscribed(currentUser, "photos", photo.ID)

		var vars = map[string]interface{}{
			"IsOwnPhoto":     currentUser.ID == user.ID,
			"User":           user,
			"Photo":          photo,
			"LikeMap":        likeMap,
			"CommentMap":     commentMap,
			"Comments":       comments,
			"CommentLikeMap": commentLikeMap,
			"IsSubscribed":   isSubscribed,
		}
		if err := tmpl.Execute(w, r, vars); err != nil {
			http.Error(w, err.Error(), http.StatusInternalServerError)
			return
		}
	})
}
